import {
  NextApiResponseWithSocket,
  cServer,
} from "../../interfaces/NextApiSocket"
import {
  composeBody,
  gameSelects,
  getAnyGame,
  getAnyRunningGame,
} from "./game/running"
import logging from "@lib/backend/logging"
import prisma from "@lib/prisma"
import colors from "colors"
import { NextApiRequest } from "next"
import { getSession } from "next-auth/react"
import { Server } from "socket.io"

colors.enable()

const SocketHandler = async (
  req: NextApiRequest,
  res: NextApiResponseWithSocket
) => {
  if (res.socket.server.io) {
    logging("Socket is already running " + req.url, ["infoCyan"], req)
  } else {
    logging("Socket is initializing " + req.url, ["infoCyan"], req)
    const io: cServer = new Server(res.socket.server, {
      path: "/api/ws",
      cors: {
        origin: "https://leaky-ships.mal-noh.de",
      },
    })

    res.socket.server.io = io

    // io.use(authenticate)
    io.use(async (socket, next) => {
      try {
        const session = await getSession({
          req: socket.request,
        })
        socket.data.user = await prisma.user.findUnique({
          where: {
            id: session?.user.id,
          },
        })

        const game = await getAnyRunningGame(socket.data.user?.id ?? "")
        if (!game) {
          logging(
            "Forbidden, no game found: " +
              JSON.stringify(Array.from(socket.rooms)),
            ["debug"],
            socket.request
          )
          return
        }
        socket.data.gameId = game.id
        socket.join(game.id)
        next()
      } catch (err) {
        logging("Unauthorized", ["warn"], socket.request)
        next(new Error("Unauthorized"))
      }
    })

    io.on("connection", async (socket) => {
      logging(
        `User connected  <${socket.data.user?.email}>`.green +
          ", " +
          socket.id.cyan,
        ["infoGreen"],
        socket.request
      )
      const game = await getAnyGame(socket.data.gameId ?? "")
      if (!socket.data.user || !socket.data.gameId || !game)
        return socket.disconnect()
      const body = composeBody(game)
      io.to(socket.data.gameId).emit("update", body)

      socket.on("gameSetting", async (payload, cb) => {
        if (!socket.data.gameId) {
          cb({ ack: false })
          return
        }
        const game = await prisma.game.update({
          where: { id: socket.data.gameId },
          data: payload,
          ...gameSelects,
        })
        const body = composeBody(game)
        cb({ ack: true })
        io.to(game.id).emit("update", body)
      })

      socket.on("ping", (count, callback) => {
        callback(count)
      })

      socket.on("disconnecting", () => {
        logging(
          "Disconnecting: " + JSON.stringify(Array.from(socket.rooms)),
          ["debug"],
          socket.request
        ) // the Set contains at least the socket ID
      })

      socket.on("disconnect", () => {
        // socket.rooms.size === 0
        logging("Disconnect: " + socket.id, ["debug"], socket.request)
      })
    })
  }
  res.end()
}

export default SocketHandler
